#include "PACXConnection.h"

#define	ONCONNEVT(evt) if(OnConnectionEvent) OnConnectionEvent(this, evt);
#define	ONDATARECVD(data) if(OnDataReceived) OnDataReceived(this, data);

#define WIN32_LEAN_AND_MEAN

#include <vector>
#include <Windows.h>
#include <WinSock2.h>
#include <WS2tcpip.h>
#include <Shlobj.h>
#include <fstream>
#include <algorithm>

#pragma comment(lib, "Ws2_32.lib")

using namespace std;
using namespace PACX;

vector<wstring> parsewstring(wstring delim, wstring input, size_t limit = 0, bool removeempty = false) {
	vector<wstring> datamap;
	wstring strtouse = input;
	size_t rc = 0;
	wstring::size_type substr_start = 0;
	wstring::size_type pos_of_delim = strtouse.find(delim);
	wstring::size_type dl = delim.length();
	while (pos_of_delim != wstring::npos) {
		if (limit > 0 && rc >= limit)
			break;
		wstring str = strtouse.substr(substr_start, pos_of_delim);
		if (str != L"" || removeempty == false)
			datamap.push_back(str);
		wstring::size_type len = strtouse.length();
		if (pos_of_delim + dl < len) {
			strtouse = strtouse.substr(pos_of_delim + dl, len - 1);
			pos_of_delim = strtouse.find(delim);
			rc++;
		}
		else if (pos_of_delim + dl == len) {
			strtouse = L"";
			if (removeempty == false)
				datamap.push_back(L"");
			rc++;
			break;
		}
		else {
			strtouse = L"";
			break;
		}
	}
	if (strtouse != L"" || (rc == 0 && removeempty == false))
		datamap.push_back(strtouse);

	return datamap;
}

bool string_ends_with(const std::wstring &str, const std::wstring &suffix) {
	return str.size() >= suffix.size() && str.compare(str.size() - suffix.size(), suffix.size(), suffix) == 0;
}

void PACXConnection::handle_input(wstring input) {
	last_received = clock();
	pending_input += input;
}

PACXConnection::PACXConnection() {
	WSADATA wsa;
	WSAStartup(MAKEWORD(2, 2), &wsa);
	connected = false;
	sock = INVALID_SOCKET;
	last_received = 0;
	last_ping = 0;
}

void PACXConnection::Connect() {
	Disconnect();
	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	int port = 9000;
	string ip = "127.0.0.1";

	PWSTR localappdata = NULL;
	SHGetKnownFolderPath(FOLDERID_LocalAppData, NULL, NULL, &localappdata);
	wstring path = wstring(localappdata) + L"\\PACX\\Settings.xml";
	wifstream ifs;
	ifs.open(path);
	if (ifs.is_open()) {
		wstring data = L"";
		wstring line = L"";
		while (getline(ifs, line)) {
			data += line;
		}
		ifs.close();
		transform(data.begin(), data.end(), data.begin(), ::towlower);

		{
			vector<wstring> ldata = parsewstring(L"<networkip>", data, 1);
			if (ldata.size() >= 2) {
				vector<wstring> ldata2 = parsewstring(L"<", ldata[1], 1);
				if (ldata2.size() >= 2) {
					ip.assign(ldata2[0].begin(), ldata2[0].end());
				}
			}
		}
		{
			vector<wstring> ldata = parsewstring(L"<networkport>", data, 1);
			if (ldata.size() >= 2) {
				vector<wstring> ldata2 = parsewstring(L"<", ldata[1], 1);
				if (ldata2.size() >= 2) {
					port = _wtoi(ldata2[0].c_str());
				}
			}
		}
	}
	if (localappdata)
		CoTaskMemFree(localappdata);

	Connect(ip, port);
}

void PACXConnection::Connect(string addr, int port) {
	sockaddr_in server;
	ZeroMemory(&server, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	inet_pton(AF_INET, addr.c_str(), &server.sin_addr);

	if (connect(sock, (sockaddr*)&server, sizeof(server)) != SOCKET_ERROR) {
		unsigned long nb = 1;
		ioctlsocket(sock, FIONBIO, (unsigned long *)&nb);

		last_received = clock();
		last_ping = clock();
		connected = true;
	}

	if (connected) {
		ONCONNEVT(PACXConnectionEventType::Connected);
	}
	else {
		ONCONNEVT(PACXConnectionEventType::ConnectionFailed);
	}
}

void PACXConnection::Send(wstring message) {
	if (!connected)
		return;


	wstring l_msg = message;
	{
		size_t matchpos = wstring::npos;
		while ((matchpos = message.find(end_sequence)) != wstring::npos) {
			message.replace(message.begin() + matchpos, message.begin() + matchpos + end_sequence.length() - 1, L"\r\n\r\n");
		}
	}


	l_msg += end_sequence;
	char* data = (char*)l_msg.c_str(); //DO NOT READ FROM THIS, for address purpose only
	int buflen = l_msg.length() * sizeof(wchar_t);
	int sent = 0;
	do {
		int lsent = send(sock, &data[sent], buflen - sent, 0);
		if (lsent >= 0) {
			sent += lsent;
		}
		else {
			Disconnect();
			break;
		}
	} while (sent < buflen);
}

void PACXConnection::Disconnect() {
	if (connected) {
		ONCONNEVT(PACXConnectionEventType::Disconnected);
	}
	if (sock != INVALID_SOCKET) {
		closesocket(sock);
		sock = INVALID_SOCKET;
	}
	last_received = 0;
	connected = false;
}

void PACXConnection::Update() {
	if (!connected)
		return;

	clock_t now = clock();
	if ((now - last_received) / CLOCKS_PER_SEC > 10.0) {
		Disconnect();
	}
	else {
		int recvd = recv(sock, buffer, 1024, 0);
		if (recvd > 0) {
			wchar_t newbuf[513];
			ZeroMemory(newbuf, sizeof(wchar_t) * 513);
			memcpy_s(newbuf, sizeof(wchar_t) * 512, buffer, recvd);
			ZeroMemory(buffer, 1024);
			handle_input(newbuf);
		}
		else if (recv == 0) {
			Disconnect();			
		}

		if (pending_input != L"") {
			vector<wstring> vals = parsewstring(end_sequence, pending_input, 0, true);
			bool use_last = string_ends_with(pending_input, end_sequence);

			size_t len = vals.size();
			for (size_t i = 0; i < len; i++) {
				if ((use_last == true || i < len - 1) && vals[i] != L"pong") {
					ONDATARECVD(vals[i]);
				}
			}
			if (use_last == false) {
				pending_input = vals[len - 1];
			}
			else {
				pending_input = L"";
			}
		}

		if ((now - last_ping) / CLOCKS_PER_SEC > 5.0) {
			last_ping = now;
			Send(L"ping");
		}
	}
}

bool PACXConnection::IsConnected() {
	return connected;
}

PACXConnection::~PACXConnection() {
	Disconnect();
	WSACleanup();
}
