#pragma once
#include <string>
#include <time.h>

namespace PACX {

	enum PACXConnectionEventType {
		Connected,
		ConnectionFailed,
		Disconnected
	};

	class PACXConnection {
	private:
		clock_t last_received;
		clock_t last_ping;
		bool connected;
		unsigned int sock;
		char buffer[1024];
		std::wstring pending_input;
		const std::wstring end_sequence = L"\r\n\r\n\\;";
		void handle_input(std::wstring input);

	public:
		typedef void(*PACXConnectionEventCallback)(PACXConnection* sender, PACXConnectionEventType eventType);
		typedef void(*PACXDataReceivedCallback)(PACXConnection* sender, std::wstring data);

		PACXConnectionEventCallback OnConnectionEvent;
		PACXDataReceivedCallback OnDataReceived;
		
		void Connect();
		void Connect(std::string addr, int port);
		void Send(std::wstring message);
		void Disconnect();

		bool IsConnected();

		void Update();

		PACXConnection();
		~PACXConnection();

	private:
		PACXConnectionEventCallback conneventcb;
	};

}
